if Code.ensure_loaded?(NimbleCSV) do
  defmodule Clickhouse.Format.TSV do
    @moduledoc """
    An implementation of the Clickhouse `"TSV"`/`"TabSeparatedValues"` format.
    """

    @behaviour Clickhouse.Format

    alias Clickhouse.{Error, Query, Result}

    @parser __MODULE__.Parser

    NimbleCSV.define(@parser, separator: "\t", escape: "\"")

    ################################
    # Clickhouse.Format Callbacks
    ################################

    @impl Clickhouse.Format
    @spec decode(Clickhouse.Result.t()) :: {:ok, Clickhouse.Result.t()}
    def decode(result) do
      data = @parser.parse_string(result.__raw__, skip_headers: false)
      result = Result.put_data(result, data)
      {:ok, result}
    end

    @impl Clickhouse.Format
    @spec encode(Clickhouse.Query.t()) ::
            {:ok, Clickhouse.Query.t()} | {:error, Clickhouse.Error.t()}
    def encode(%{params: []} = query) do
      {:ok, query}
    end

    def encode(%{params_count: 0} = query) do
      {:ok, query}
    end

    def encode(%{type: :insert, params_count: count}) when count > 1 do
      error = Error.new(:invalid_statement, "TSV insert statement can only support a single '?'.")
      {:error, error}
    end

    def encode(%{type: :insert, params: params, params_count: 1} = query) do
      data = params |> do_encode() |> @parser.dump_to_iodata() |> IO.iodata_to_binary()
      encoded = String.replace(query.statement, "?", data)
      {:ok, Query.put_statement(query, encoded)}
    end

    ################################
    # Private API
    ################################

    defp do_encode(params), do: Enum.map(params, &encode_row/1)

    defp encode_row(values), do: Enum.map(values, &encode_value/1)

    defp encode_value(nil), do: nil
    defp encode_value(string) when is_binary(string), do: "'" <> string <> "'"
    defp encode_value(integer) when is_integer(integer), do: to_string(integer)
    defp encode_value(atom) when is_atom(atom), do: to_string(atom)
    defp encode_value(false), do: 0
    defp encode_value(true), do: 1

    defp encode_value(list) when is_list(list) do
      values = Enum.map(list, &encode_value/1)
      "[" <> Enum.join(values, ",") <> "]"
    end
  end
end
